from rapidata.rapidata_client.settings import (
    AlertOnFastResponse,
    TranslationBehaviour,
    FreeTextMinimumCharacters,
    NoShuffle,
    PlayVideoUntilTheEnd,
    CustomSetting,
    )

class RapidataSettings:
    """
    Container class for all setting factory functions

    Settings can be added to an order to determine the behaviour of the task.
    
    Attributes:
        alert_on_fast_response (AlertOnFastResponse): The AlertOnFastResponse instance.
        translation_behaviour (TranslationBehaviour): The TranslationBehaviour instance.
        free_text_minimum_characters (FreeTextMinimumCharacters): The FreeTextMinimumCharacters instance.
        no_shuffle (NoShuffle): The NoShuffle instance.
        play_video_until_the_end (PlayVideoUntilTheEnd): The PlayVideoUntilTheEnd instance.
        custom_setting (CustomSetting): The CustomSetting instance.
    """

    alert_on_fast_response = AlertOnFastResponse
    translation_behaviour = TranslationBehaviour
    free_text_minimum_characters = FreeTextMinimumCharacters
    no_shuffle = NoShuffle
    play_video_until_the_end = PlayVideoUntilTheEnd
    custom_setting = CustomSetting




from rapidata.rapidata_client.assets.data_type_enum import RapidataDataTypes
from rapidata.rapidata_client.validation.rapids.rapids import (
    ClassificationRapid, 
    CompareRapid, 
    SelectWordsRapid, 
    LocateRapid, 
    DrawRapid)
from rapidata.rapidata_client.assets import MediaAsset, TextAsset, MultiAsset
from rapidata.rapidata_client.metadata import Metadata
from rapidata.rapidata_client.validation.rapids.box import Box

from typing import Sequence

class RapidsManager:
    """
    Can be used to build different types of rapids. That can then be added to Validation sets
    """
    def __init__(self):
        pass
    
    def classification_rapid(self,
            instruction: str,
            answer_options: list[str],
            datapoint: str,
            truths: list[str],
            data_type: str = RapidataDataTypes.MEDIA,
            metadata: Sequence[Metadata] = [],
    ) -> ClassificationRapid:
        """Build a classification rapid
        
        Args:
            instruction (str): The instruction/question to be shown to the labeler.
            answer_options (list[str]): The options that the labeler can choose from to answer the question.
            datapoint (str): The datapoint that the labeler will be labeling.
            truths (list[str]): The correct answers to the question.
            data_type (str, optional): The type of the datapoint. Defaults to RapidataDataTypes.MEDIA.
            metadata (Sequence[Metadata], optional): The metadata that is attached to the rapid. Defaults to [].
        """
        
        if data_type == RapidataDataTypes.MEDIA:
            asset = MediaAsset(datapoint)
        elif data_type == RapidataDataTypes.TEXT:
            asset = TextAsset(datapoint)
        else:
            raise ValueError(f"Unsupported data type: {data_type}")

        return ClassificationRapid(
                instruction=instruction,
                answer_options=answer_options,
                asset=asset,
                truths=truths,
                metadata=metadata,
                )
    
    def compare_rapid(self,
            instruction: str,
            truth: str,
            datapoint: list[str],
            data_type: str = RapidataDataTypes.MEDIA,
            metadata: Sequence[Metadata] = [],
    ) -> CompareRapid:
        """Build a compare rapid

        Args:
            instruction (str): The instruction that the labeler will be comparing the assets on.
            truth (str): The correct answer to the comparison. (has to be one of the assets)
            datapoint (list[str]): The two assets that the labeler will be comparing.
            data_type (str, optional): The type of the datapoint. Defaults to RapidataDataTypes.MEDIA.
            metadata (Sequence[Metadata], optional): The metadata that is attached to the rapid. Defaults to [].
        """

        if data_type == RapidataDataTypes.MEDIA:
            assets = [MediaAsset(image) for image in datapoint]
        elif data_type == RapidataDataTypes.TEXT:
            assets = [TextAsset(text) for text in datapoint]
        else:
            raise ValueError(f"Unsupported data type: {data_type}")
        
        asset = MultiAsset(assets)

        return CompareRapid(
                instruction=instruction,
                asset=asset,
                truth=truth,
                metadata=metadata,
                )
    
    def select_words_rapid(self,
            instruction: str,
            truths: list[int],
            datapoint: str,
            sentence: str,
            strict_grading: bool = True,
    ) -> SelectWordsRapid:
        """Build a select words rapid

        Args:
            instruction (str): The instruction for the labeler.
            truths (list[int]): The indices of the words that are the correct answers.
            datapoint (str): The asset that the labeler will be selecting words from.
            sentence (str): The sentence that the labeler will be selecting words from. (split up by spaces)
            strict_grading (bool, optional): Whether the grading should be strict or not. 
                True means that all correct words and no wrong words have to be selected for the rapid to be marked as correct.
                False means that at least one correct word and no wrong words have to be selected for the rapid to be marked as correct. Defaults to True.
        """
        
        asset = MediaAsset(datapoint)

        return SelectWordsRapid(
                instruction=instruction,
                truths=truths,
                asset=asset,
                sentence=sentence,
                strict_grading=strict_grading,
                )
    
    def locate_rapid(self,
            instruction: str,
            truths: list[Box],
            datapoint: str,
            metadata: Sequence[Metadata] = [],
    ) -> LocateRapid:
        """Build a locate rapid

        Args:
            instruction (str): The instruction on what the labeler should do.
            truths (list[Box]): The bounding boxes of the object that the labeler ought to be locating.
            datapoint (str): The asset that the labeler will be locating the object in.
            metadata (Sequence[Metadata], optional): The metadata that is attached to the rapid. Defaults to [].
        """
        
        asset = MediaAsset(datapoint)

        return LocateRapid(
                instruction=instruction,
                truths=truths,
                asset=asset,
                metadata=metadata,
                )
    
    def draw_rapid(self,
            instruction: str,
            truths: list[Box],
            datapoint: str,
            metadata: Sequence[Metadata] = [],
    ) -> DrawRapid:
        """Build a draw rapid

        Args:
            instruction (str): The instructions on what the labeler
            truths (list[Box]): The bounding boxes of the object that the labeler ought to be drawing.
            datapoint (str): The asset that the labeler will be drawing the object in.
            metadata (Sequence[Metadata], optional): The metadata that is attached to the rapid. Defaults to [].
        """
        
        asset = MediaAsset(datapoint)

        return DrawRapid(
                instruction=instruction,
                truths=truths,
                asset=asset,
                metadata=metadata,
                )



from rapidata.api_client.models.capped_selection_selections_inner import (
    CappedSelectionSelectionsInner,
)
from rapidata.api_client.models.create_order_model import CreateOrderModel
from rapidata.api_client.models.create_order_model_referee import (
    CreateOrderModelReferee,
)
from rapidata.api_client.models.create_order_model_user_filters_inner import (
    CreateOrderModelUserFiltersInner,
)
from rapidata.api_client.models.create_order_model_workflow import (
    CreateOrderModelWorkflow,
)

from rapidata.rapidata_client.settings import RapidataSetting
from rapidata.rapidata_client.metadata._base_metadata import Metadata
from rapidata.rapidata_client.order._rapidata_dataset import RapidataDataset
from rapidata.rapidata_client.referee._naive_referee import NaiveReferee
from rapidata.rapidata_client.selection._base_selection import RapidataSelection
from rapidata.rapidata_client.filter import RapidataFilter
from rapidata.rapidata_client.workflow import Workflow
from rapidata.rapidata_client.order.rapidata_order import RapidataOrder
from rapidata.rapidata_client.referee import Referee
from rapidata.service.openapi_service import OpenAPIService

from rapidata.rapidata_client.workflow._compare_workflow import CompareWorkflow

from rapidata.rapidata_client.assets import MediaAsset, TextAsset, MultiAsset, BaseAsset

from typing import Optional, cast, Sequence


class RapidataOrderBuilder:
    """Builder object for creating Rapidata orders.

    Use the fluent interface to set the desired configuration. Add a workflow to the order using `.workflow()` and finally call `.create()` to create the order.

    Args:
        openapi_service (OpenAPIService): The OpenAPIService instance.
        name (str): The name of the order.
    """

    def __init__(
        self,
        name: str,
        openapi_service: OpenAPIService,
    ):
        self._name = name
        self.order_id: str | None = None
        self.__openapi_service = openapi_service
        self.__dataset: Optional[RapidataDataset]
        self.__workflow: Workflow | None = None
        self.__referee: Referee | None = None
        self.__metadata: Sequence[Metadata] | None = None
        self.__validation_set_id: str | None = None
        self.__settings: Sequence[RapidataSetting] | None = None
        self.__user_filters: list[RapidataFilter] = []
        self.__selections: list[RapidataSelection] = []
        self.__priority: int = 50
        self.__assets: Sequence[BaseAsset] = []

    def _to_model(self) -> CreateOrderModel:
        """
        Convert the builder configuration to a CreateOrderModel.

        Raises:
            ValueError: If no workflow is provided.

        Returns:
            CreateOrderModel: The model representing the order configuration.
        """
        if self.__workflow is None:
            raise ValueError("You must provide a workflow to create an order.")

        if self.__referee is None:
            print("No referee provided, using default NaiveReferee.")
            self.__referee = NaiveReferee()

        return CreateOrderModel(
            _t="CreateOrderModel",
            orderName=self._name,
            workflow=CreateOrderModelWorkflow(self.__workflow._to_model()),
            userFilters=[
                CreateOrderModelUserFiltersInner(user_filter._to_model())
                for user_filter in self.__user_filters
            ],
            referee=CreateOrderModelReferee(self.__referee._to_model()),
            validationSetId=self.__validation_set_id,
            featureFlags=(
                [setting._to_feature_flag() for setting in self.__settings]
                if self.__settings is not None
                else None
            ),
            selections=[
                CappedSelectionSelectionsInner(selection._to_model())
                for selection in self.__selections
            ],
            priority=self.__priority,
        )

    def _create(self, max_upload_workers: int = 10) -> RapidataOrder:
        """
        Create the Rapidata order by making the necessary API calls based on the builder's configuration.

        Args:
            max_upload_workers (int, optional): The maximum number of worker threads for processing media paths. Defaults to 10.

        Raises:
            ValueError: If both media paths and texts are provided, or if neither is provided.
            AssertionError: If the workflow is a CompareWorkflow and media paths are not in pairs.

        Returns:
            RapidataOrder: The created RapidataOrder instance.
        """
        order_model = self._to_model()
        if isinstance(
            self.__workflow, CompareWorkflow
        ):  # Temporary fix; will be handled by backend in the future
            assert all(
                isinstance(item, MultiAsset) for item in self.__assets
            ), "The media paths must be of type MultiAsset for comparison tasks."

        result = self.__openapi_service.order_api.order_create_post(
            create_order_model=order_model
        )

        self.order_id = result.order_id

        self.__dataset = (
            RapidataDataset(result.dataset_id, self.__openapi_service)
            if result.dataset_id
            else None
        )

        order = RapidataOrder(
            order_id=self.order_id,
            dataset=self.__dataset,
            openapi_service=self.__openapi_service,
            name=self._name,
        )

        if all(isinstance(item, MediaAsset) for item in self.__assets) and order.dataset:
            assets = cast(list[MediaAsset], self.__assets)
            order.dataset._add_media_from_paths(assets, self.__metadata, max_upload_workers)

        elif (
            all(isinstance(item, TextAsset) for item in self.__assets) and order.dataset
        ):
            assets = cast(list[TextAsset], self.__assets)
            order.dataset._add_texts(assets)

        elif (
            all(isinstance(item, MultiAsset) for item in self.__assets) and order.dataset
        ):
            multi_assets = cast(list[MultiAsset], self.__assets)

            # Check if all MultiAssets contain the same type of assets
            first_asset_type = type(multi_assets[0].assets[0])
            if not all(
                isinstance(asset, first_asset_type)
                for multi_asset in multi_assets
                for asset in multi_asset.assets
            ):
                raise ValueError(
                    "All MultiAssets must contain the same type of assets (either all MediaAssets or all TextAssets)."
                )

            # Process based on the asset type
            if issubclass(first_asset_type, MediaAsset):
                order.dataset._add_media_from_paths(
                    multi_assets, self.__metadata, max_upload_workers
                )

            elif issubclass(first_asset_type, TextAsset):
                order.dataset._add_texts(multi_assets)

            else:
                raise ValueError(
                    "MultiAsset must contain MediaAssets or TextAssets objects."
                )

        elif order.dataset:
            raise ValueError(
                "Media paths must all be of the same type: MediaAsset, TextAsset, or MultiAsset."
            )

        return order

    def _workflow(self, workflow: Workflow) -> "RapidataOrderBuilder":
        """
        Set the workflow for the order.

        Args:
            workflow (Workflow): The workflow to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(workflow, Workflow):
            raise TypeError("Workflow must be of type Workflow.")

        self.__workflow = workflow
        return self

    def _referee(self, referee: Referee) -> "RapidataOrderBuilder":
        """
        Set the referee for the order.

        Args:
            referee (Referee): The referee to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(referee, Referee):
            raise TypeError("Referee must be of type Referee.")

        self.__referee = referee
        return self

    def _media(
        self,
        asset: Sequence[BaseAsset],
        metadata: Sequence[Metadata] | None = None,
    ) -> "RapidataOrderBuilder":
        """
        Set the media assets for the order.

        Args:
            asset: (list[MediaAsset] | list[TextAsset] | list[MultiAsset]): The paths of the media assets to be set.
            metadata: (list[Metadata] | None, optional): Metadata for the media assets. Defaults to None.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(asset, list):
            raise TypeError("Media paths must be provided as a list of paths.")

        # for a in asset:
        #     if not isinstance(a, (MediaAsset, TextAsset, MultiAsset)):
        #         raise TypeError(
        #             "Media paths must be of type MediaAsset, TextAsset, or MultiAsset."
        #         )

        if metadata:
            for data in metadata:
                if not isinstance(data, Metadata):
                    raise TypeError("Metadata must be of type Metadata.")

        self.__assets = asset
        self.__metadata = metadata
        return self

    def _settings(self, settings: Sequence[RapidataSetting]) -> "RapidataOrderBuilder":
        """
        Set the settings for the order.

        Args:
            settings (Settings): The settings to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """

        if not isinstance(settings, list):
            raise TypeError("Settings must be provided as a list of Setting objects.")
        
        for s in settings:
            if not isinstance(s, RapidataSetting):
                raise TypeError("The settings list must only contain Setting objects.")

        self.__settings = settings
        return self

    def _filters(self, filters: Sequence[RapidataFilter]) -> "RapidataOrderBuilder":
        """
        Set the filters for the order, e.g., country, language, userscore, etc.

        Args:
            filters (Sequence[Filters]): The user filters to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(filters, list):
            raise TypeError("Filters must be provided as a list of Filter objects.")

        for f in filters:
            if not isinstance(f, RapidataFilter):
                raise TypeError("Filters must be of type Filter.")

        if len(self.__user_filters) > 0:
            print("Overwriting existing user filters.")

        self.__user_filters = filters
        return self

    def _validation_set_id(self, validation_set_id: str) -> "RapidataOrderBuilder":
        """
        Set the validation set ID for the order.

        Args:
            validation_set_id (str): The validation set ID to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(validation_set_id, str):
            raise TypeError("Validation set ID must be of type str.")

        self.__validation_set_id = validation_set_id
        return self

    def _rapids_per_bag(self, amount: int) -> "RapidataOrderBuilder":
        """
        Define the number of tasks a user sees in a single session.

        Args:
            amount (int): The number of tasks a user sees in a single session.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.

        Raises:
            NotImplementedError: This method is not implemented yet.
        """
        raise NotImplementedError("Not implemented yet.")

    def _selections(self, selections: Sequence[RapidataSelection]) -> "RapidataOrderBuilder":
        """
        Set the selections for the order.

        Args:
            selections (Sequence[Selection]): The selections to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(selections, list):
            raise TypeError(
                "Selections must be provided as a list of Selection objects."
            )

        for selection in selections:
            if not isinstance(selection, RapidataSelection):
                raise TypeError("Selections must be of type Selection.")

        self.__selections = selections  # type: ignore
        return self

    def _priority(self, priority: int) -> "RapidataOrderBuilder":
        """
        Set the priority for the order.

        Args:
            priority (int): The priority to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(priority, int):
            raise TypeError("Priority must be of type int.")

        self.__priority = priority
        return self



from rapidata.rapidata_client.filter import (
    AgeFilter, 
    CampaignFilter, 
    CountryFilter, 
    GenderFilter, 
    LanguageFilter, 
    UserScoreFilter,
    CustomFilter)

class RapidataFilters:
    """RapidataFilters Classes

    These filters can be added to the order to specifically target a certain group of users.

    Note that adding multiple filters to the same order will result in a logical AND operation between the filters.

    Warning: this might significantly slow down the number of responses you receive.
    
    Attributes:
        user_score (UserScoreFilter): The UserScoreFilter instance.
        age (AgeFilter): The AgeFilter instance.
        campaign (CampaignFilter): The CampaignFilter instance.
        country (CountryFilter): The CountryFilter instance.
        campaign (CampaignFilter): The CampaignFilter instance
        gender (GenderFilter): The GenderFilter instance.
        language (LanguageFilter): The LanguageFilter instance.
        custom_filter (CustomFilter): The CustomFilter instance.
    """
    user_score = UserScoreFilter
    age = AgeFilter 
    campaign = CampaignFilter
    country = CountryFilter
    gender = GenderFilter
    language = LanguageFilter
    custom_filter = CustomFilter



from rapidata.rapidata_client.validation._validation_set_builder import ValidationSetBuilder
from rapidata.rapidata_client.validation.rapidata_validation_set import RapidataValidationSet
from rapidata.service.openapi_service import OpenAPIService
from rapidata.rapidata_client.assets.data_type_enum import RapidataDataTypes
from rapidata.rapidata_client.validation.rapids.rapids_manager import RapidsManager
from rapidata.rapidata_client.validation.rapids.rapids import Rapid
from rapidata.rapidata_client.metadata import PromptMetadata

from rapidata.api_client.models.page_info import PageInfo
from rapidata.api_client.models.root_filter import RootFilter
from rapidata.api_client.models.filter import Filter
from rapidata.api_client.models.sort_criterion import SortCriterion
from rapidata.api_client.exceptions import BadRequestException
from urllib3._collections import HTTPHeaderDict

from rapidata.rapidata_client.validation.rapids.box import Box

from rapidata.api_client.models.query_validation_set_model import QueryValidationSetModel

from typing import Sequence


class ValidationSetManager:
    """
    Responsible for everything related to validation sets. From creation to retrieval.

    Attributes:
        rapid (RapidsManager): The RapidsManager instance.
    """
    def __init__(self, openapi_service: OpenAPIService) -> None:
        self.__openapi_service = openapi_service
        self.rapid = RapidsManager()

    def create_classification_set(self,
        name: str,
        instruction: str,
        answer_options: list[str],
        datapoints: list[str],
        truths: list[list[str]],
        data_type: str = RapidataDataTypes.MEDIA,
        contexts: list[str] | None = None,
        print_confirmation: bool = True
    ) -> RapidataValidationSet:
        """Create a classification validation set.
        
        Args:
            name (str): The name of the validation set. (will not be shown to the labeler)
            instruction (str): The instruction by which the labeler will answer.
            answer_options (list[str]): The options to choose from when answering.
            datapoints (list[str]): The datapoints that will be used for validation.
            truths (list[list[str]]): The truths for each datapoint. Outher list is for each datapoint, inner list is for each truth.\n
                example:
                    options: ["yes", "no", "maybe"]
                    datapoints: ["datapoint1", "datapoint2"]
                    truths: [["yes"], ["no", "maybe"]] -> first datapoint correct answer is "yes", second datapoint is "no" or "maybe"
            data_type (str, optional): The type of data. Defaults to RapidataDataTypes.MEDIA. Other option: RapidataDataTypes.TEXT ("text").
            contexts (list[str], optional): The contexts for each datapoint. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction and answer options. (Therefore will be different for each datapoint)
                Will be match up with the datapoints using the list index.
            print_confirmation (bool, optional): Whether to print a confirmation message that validation set has been created. Defaults to True.
        """
        
        if len(datapoints) != len(truths):
            raise ValueError("The number of datapoints and truths must be equal")
        
        if not all([isinstance(truth, list) for truth in truths]):
            raise ValueError("Truths must be a list of lists")
        
        if contexts and len(contexts) != len(datapoints):
            raise ValueError("The number of contexts and datapoints must be equal")
        
        rapids = []
        for i in range(len(datapoints)):
            rapids.append(
                self.rapid.classification_rapid(
                    instruction=instruction,
                    answer_options=answer_options,
                    datapoint=datapoints[i],
                    truths=truths[i],
                    data_type=data_type,
                    metadata=[PromptMetadata(contexts[i])] if contexts else []
                )
            )

        validation_set_builder = ValidationSetBuilder(name, self.__openapi_service)
        for rapid in rapids:
            validation_set_builder._add_rapid(rapid)

        return validation_set_builder._submit(print_confirmation)
    
    def create_compare_set(self,
        name: str,
        instruction: str,
        datapoints: list[list[str]],
        truths: list[str],
        data_type: str = RapidataDataTypes.MEDIA,
        contexts: list[str] | None = None,
        print_confirmation: bool = True
    ) -> RapidataValidationSet:
        """Create a comparison validation set.

        Args:
            name (str): The name of the validation set. (will not be shown to the labeler)
            instruction (str): The instruction to compare against.
            truths (list[str]): The truths for each comparison. List is for each comparison.\n
                example:
                    instruction: "Which image has a cat?"
                    datapoints = [["image1.jpg", "image2.jpg"], ["image3.jpg", "image4.jpg"]]
                    truths: ["image1.jpg", "image4.jpg"] -> first comparison image1.jpg has a cat, second comparison image4.jpg has a cat
            datapoints (list[list[str]]): The compare datapoints to create the validation set with. 
                Outer list is for each comparison, inner list the two images/texts that will be compared.
            data_type (str, optional): The type of data. Defaults to RapidataDataTypes.MEDIA. Other option: RapidataDataTypes.TEXT ("text").
            contexts (list[str], optional): The contexts for each datapoint. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction and truth. (Therefore will be different for each datapoint)
                Will be match up with the datapoints using the list index.
            print_confirmation (bool, optional): Whether to print a confirmation message that validation set has been created. Defaults to True.
        """
        
        if len(datapoints) != len(truths):
            raise ValueError("The number of datapoints and truths must be equal")
        
        if not all([isinstance(truth, list) for truth in truths]):
            raise ValueError("Truths must be a list of lists")

        if contexts and len(contexts) != len(datapoints):
            raise ValueError("The number of contexts and datapoints must be equal")
        
        rapids = []
        for i in range(len(datapoints)):
            rapids.append(
                self.rapid.compare_rapid(
                    instruction=instruction,
                    truth=truths[i],
                    datapoint=datapoints[i],
                    data_type=data_type,
                    metadata=[PromptMetadata(contexts[i])] if contexts else []
                )
            )

        validation_set_builder = ValidationSetBuilder(name, self.__openapi_service)
        for rapid in rapids:
            validation_set_builder._add_rapid(rapid)

        return validation_set_builder._submit(print_confirmation)
    
    def create_select_words_set(self,
        name: str,
        instruction: str,
        truths: list[list[int]],
        datapoints: list[str],
        sentences: list[str],
        strict_grading: bool = True,
        print_confirmation: bool = True
    ) -> RapidataValidationSet:
        """Create a select words validation set.

        Args:
            name (str): The name of the validation set. (will not be shown to the labeler)
            instruction (str): The instruction to show to the labeler.
            truths (list[list[int]]): The truths for each datapoint. Outher list is for each datapoint, inner list is for each truth.\n
                example:
                    datapoints: ["datapoint1", "datapoint2"]
                    sentences: ["this example 1", "this example 2"]
                    truths: [[0, 1], [2]] -> first datapoint correct words are "this" and "example", second datapoint is "2"
            datapoints (list[str]): The datapoints that will be used for validation.
            sentences (list[str]): The sentences that will be used for validation. The sentece will be split up by spaces to be selected by the labeler.
                Must be the same length as datapoints.
            strict_grading (bool, optional): Whether to grade strictly. Defaults to True.\n
                If True, the labeler must select all correct words to be graded as correct. (and no wrong words)
                If False, the labeler must select at least one correct word to be graded as correct. (and no wrong words)
            print_confirmation (bool, optional): Whether to print a confirmation message that validation set has been created. Defaults to True.
            """
        
        if not all([isinstance(truth, list) for truth in truths]):
            raise ValueError("Truths must be a list of lists")

        if len(datapoints) != len(truths) or len(datapoints) != len(sentences):
            raise ValueError("The number of datapoints, truths, and sentences must be equal")
        
        rapids = []
        for i in range(len(datapoints)):
            rapids.append(
                self.rapid.select_words_rapid(
                    instruction=instruction,
                    truths=truths[i],
                    datapoint=datapoints[i],
                    sentence=sentences[i],
                    strict_grading=strict_grading
                )
            )

        validation_set_builder = ValidationSetBuilder(name, self.__openapi_service)
        for rapid in rapids:
            validation_set_builder._add_rapid(rapid)

        return validation_set_builder._submit(print_confirmation)
    
    def create_locate_set(self,
        name: str,
        instruction: str,
        truths: list[list[Box]],
        datapoints: list[str],
        contexts: list[str] | None = None,
        print_confirmation: bool = True
    ) -> RapidataValidationSet:
        """Create a locate validation set.

        Args:
            name (str): The name of the validation set. (will not be shown to the labeler)
            instruction (str): The instruction to show to the labeler.
            truths (list[list[Box]]): The truths for each datapoint. Outher list is for each datapoint, inner list is for each truth.\n
                example:
                    datapoints: ["datapoint1", "datapoint2"]
                    truths: [[Box(0, 0, 100, 100)], [Box(50, 50, 150, 150)]] -> first datapoint the object is in the top left corner, second datapoint the object is in the center
            datapoints (list[str]): The datapoints that will be used for validation.
            contexts (list[str], optional): The contexts for each datapoint. Defaults to None.
            print_confirmation (bool, optional): Whether to print a confirmation message that validation set has been created. Defaults to True.
        """
        
        if len(datapoints) != len(truths):
            raise ValueError("The number of datapoints and truths must be equal")
        
        if not all([isinstance(truth, list) for truth in truths]):
            raise ValueError("Truths must be a list of lists")
        
        if contexts and len(contexts) != len(datapoints):
            raise ValueError("The number of contexts and datapoints must be equal")
        
        rapids = []
        for i in range(len(datapoints)):
            rapids.append(
                self.rapid.locate_rapid(
                    instruction=instruction,
                    truths=truths[i],
                    datapoint=datapoints[i],
                    metadata=[PromptMetadata(contexts[i])] if contexts else []
                )
            )

        validation_set_builder = ValidationSetBuilder(name, self.__openapi_service)
        for rapid in rapids:
            validation_set_builder._add_rapid(rapid)

        return validation_set_builder._submit(print_confirmation)
    
    def create_draw_set(self,
        name: str,
        instruction: str,
        truths: list[list[Box]],
        datapoints: list[str],
        contexts: list[str] | None = None,
        print_confirmation: bool = True
    ) -> RapidataValidationSet:
        """Create a draw validation set.

        Args:
            name (str): The name of the validation set. (will not be shown to the labeler)
            instruction (str): The instruction to show to the labeler.
            truths (list[list[Box]]): The truths for each datapoint. Outher list is for each datapoint, inner list is for each truth.\n
                example:
                    datapoints: ["datapoint1", "datapoint2"]
                    truths: [[Box(0, 0, 100, 100)], [Box(50, 50, 150, 150)]] -> first datapoint the object is in the top left corner, second datapoint the object is in the center
            datapoints (list[str]): The datapoints that will be used for validation.
            contexts (list[str], optional): The contexts for each datapoint. Defaults to None.
            print_confirmation (bool, optional): Whether to print a confirmation message that validation set has been created. Defaults to True.
        """
        
        if len(datapoints) != len(truths):
            raise ValueError("The number of datapoints and truths must be equal")
        
        if not all([isinstance(truth, list) for truth in truths]):
            raise ValueError("Truths must be a list of lists")
        
        if contexts and len(contexts) != len(datapoints):
            raise ValueError("The number of contexts and datapoints must be equal")
        
        rapids = []
        for i in range(len(datapoints)):
            rapids.append(
                self.rapid.draw_rapid(
                    instruction=instruction,
                    truths=truths[i],
                    datapoint=datapoints[i],
                    metadata=[PromptMetadata(contexts[i])] if contexts else []
                )
            )

        validation_set_builder = ValidationSetBuilder(name, self.__openapi_service)
        for rapid in rapids:
            validation_set_builder._add_rapid(rapid)

        return validation_set_builder._submit(print_confirmation)
    
    def create_mixed_set(self,
        name: str,
        rapids: Sequence[Rapid],
        print_confirmation: bool = True
    ) -> RapidataValidationSet:
        """Create a validation set with a list of rapids.

        Args:
            name (str): The name of the validation set. (will not be shown to the labeler)
            rapids (Sequence[Rapid]): The list of rapids to add to the validation set.
            print_confirmation (bool, optional): Whether to print a confirmation message that validation set has been created. Defaults to True.
        """

        validation_set_builder = ValidationSetBuilder(name, self.__openapi_service)
        for rapid in rapids:
            validation_set_builder._add_rapid(rapid)

        return validation_set_builder._submit(print_confirmation)
    
    def get_validation_set_by_id(self, validation_set_id: str) -> RapidataValidationSet:
        """Get a validation set by ID.

        Args:
            validation_set_id (str): The ID of the validation set.

        Returns:
            RapidataValidationSet: The ValidationSet instance.
        """
        try:
            validation_set = self.__openapi_service.validation_api.validation_get_by_id_get(id=validation_set_id)
        except Exception:
            raise ValueError(f"ValidationSet with ID {validation_set_id} not found.")
        
        return RapidataValidationSet(validation_set_id, self.__openapi_service, validation_set.name)
    
    def find_validation_sets(self, name: str = "", amount: int = 1) -> list[RapidataValidationSet]:
        """Find validation sets by name.

        Args:
            name (str, optional): The name to search for. Defaults to "" to match with any set.
            amount (int, optional): The amount of validation sets to return. Defaults to 1.

        Returns:
            list[RapidataValidationSet]: The list of validation sets.
        """
        try:
            validation_page_result = self.__openapi_service.validation_api.validation_query_validation_sets_get(QueryValidationSetModel(
                pageInfo=PageInfo(index=1, size=amount),
                filter=RootFilter(filters=[Filter(field="Name", operator="Contains", value=name)]),
                sortCriteria=[SortCriterion(direction="Desc", propertyName="CreatedAt")]
                ))

        except BadRequestException as e:
            raise ValueError(f"Error occured during request. \nError: {e.body} \nTraceid: {e.headers.get('X-Trace-Id') if isinstance(e.headers, HTTPHeaderDict) else 'Unknown'}")

        except Exception as e:
            raise ValueError(f"Unknown error occured: {e}")

        validation_sets = [self.get_validation_set_by_id(validation_set.id) for validation_set in validation_page_result.items]
        return validation_sets




from ._base_filter import RapidataFilter
from .age_filter import AgeFilter, AgeGroup
from .campaign_filter import CampaignFilter
from .country_filter import CountryFilter
from .gender_filter import GenderFilter, Gender
from .language_filter import LanguageFilter
from .user_score_filter import UserScoreFilter
from .custom_filter import CustomFilter



from enum import Enum

class TranslationBehaviourOptions(Enum):
    """The options for the translation behaviour setting.
    
    Attributes:
        BOTH: Show both the original and the translated text.
            May clutter the screen if the options are too long.
        ONLY_ORIGINAL: Show only the original text.
        ONLY_TRANSLATED: Show only the translated text."""
    
    BOTH = "both"
    ONLY_ORIGINAL = "only original"
    ONLY_TRANSLATED = "only translated"



from rapidata.service.openapi_service import OpenAPIService

from rapidata.rapidata_client.order.rapidata_order_manager import RapidataOrderManager

from rapidata.rapidata_client.validation.validation_set_manager import ValidationSetManager


class RapidataClient:
    """The Rapidata client is the main entry point for interacting with the Rapidata API. It allows you to create orders and validation sets."""
    
    def __init__(
        self,
        client_id: str | None = None,
        client_secret: str | None = None,
        enviroment: str = "rapidata.ai",
        oauth_scope: str = "openid",
        cert_path: str | None = None,
    ):
        """Initialize the RapidataClient. If both the client_id and client_secret are None, it will try using your credentials under "~/.config/rapidata/credentials.json". 
        If this is not successful, it will open a browser windown and ask you to log in, then save your new credentials in said json file.

        Args:
            client_id (str): The client ID for authentication.
            client_secret (str): The client secret for authentication.
            enviroment (str, optional): The API endpoint.

        Attributes:
            order (RapidataOrderManager): The RapidataOrderManager instance.
            validation (ValidationSetManager): The ValidationSetManager instance.
        """
        self._openapi_service = OpenAPIService(
            client_id=client_id,
            client_secret=client_secret,
            enviroment=enviroment,
            oauth_scope=oauth_scope,
            cert_path=cert_path
        )
        
        self.order = RapidataOrderManager(openapi_service=self._openapi_service)
        
        self.validation = ValidationSetManager(openapi_service=self._openapi_service)



from rapidata.rapidata_client.selection import (
    DemographicSelection, 
    LabelingSelection, 
    ValidationSelection, 
    ConditionalValidationSelection, 
    CappedSelection)

class RapidataSelections:
    """RapidataSelections Classes

    Attributes:
        demographic (DemographicSelection): The DemographicSelection instance.
        labeling (LabelingSelection): The LabelingSelection instance.
        validation (ValidationSelection): The ValidationSelection instance.
        conditional_validation (ConditionalValidationSelection): The ConditionalValidationSelection instance.
        capped (CappedSelection): The CappedSelection instance."""
    demographic = DemographicSelection
    labeling = LabelingSelection
    validation = ValidationSelection
    conditional_validation = ConditionalValidationSelection
    capped = CappedSelection



from rapidata.rapidata_client.selection._base_selection import RapidataSelection
from rapidata.api_client.models.conditional_validation_rapid_selection_config import (
    ValidationChance,
)
from rapidata.api_client.models.conditional_validation_selection import (
    ConditionalValidationSelection as ConditionalValidationSelectionModel,
)


class ConditionalValidationSelection(RapidataSelection):
    """Conditional validation selection class.

    Probabilistically decides how many validation rapids you want to show per session based on the user score.
    
    Args:
        validation_set_id (str): The id of the validation set to be used.
        thresholds (list[float]): The thresholds to use for the user score.
        chances (list[float]): The chances of showing a validation rapid for each threshold.
        rapid_counts (list[int]): The amount of validation rapids that will be shown per session of this validation set for each threshold if selected by probability. (all or nothing)
    """

    def __init__(
        self,
        validation_set_id: str,
        thresholds: list[float],
        chances: list[float],
        rapid_counts: list[int],
    ):
        if len(thresholds) != len(chances) or len(thresholds) != len(rapid_counts):
            raise ValueError(
                "The lengths of thresholds, chances and rapid_counts must be equal."
            )
        self.validation_set_id = validation_set_id
        self.thresholds = thresholds
        self.chances = chances
        self.rapid_counts = rapid_counts

    def _to_model(self):
        return ConditionalValidationSelectionModel(
            _t="ConditionalValidationSelection",
            validationSetId=self.validation_set_id,
            validationChances=[
                ValidationChance(
                    userScoreThreshold=threshold, chance=chance, rapidCount=rapid_count
                )
                for threshold, chance, rapid_count in zip(
                    self.thresholds, self.chances, self.rapid_counts
                )
            ],
        )



from rapidata.rapidata_client.settings._rapidata_setting import RapidataSetting

class FreeTextMinimumCharacters(RapidataSetting):
    """
    Set the minimum number of characters a user has to type.
    
    Args:
        value (int): The minimum number of characters for free text.
    """

    def __init__(self, value: int):
        if value < 1:
            raise ValueError("The minimum number of characters must be greater than or equal to 1.")
        if value > 40:
            print(f"Warning: Are you sure you want to set the minimum number of characters at {value}?")
        super().__init__(key="free_text_minimum_characters", value=value)



# coding: utf-8

"""
    Rapidata.Dataset

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: v1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import json
import pprint
from pydantic import BaseModel, ConfigDict, Field, StrictStr, ValidationError, field_validator
from typing import Any, List, Optional
from rapidata.api_client.models.attach_category_rapid_blueprint import AttachCategoryRapidBlueprint
from rapidata.api_client.models.bounding_box_rapid_blueprint import BoundingBoxRapidBlueprint
from rapidata.api_client.models.compare_rapid_blueprint import CompareRapidBlueprint
from rapidata.api_client.models.free_text_rapid_blueprint import FreeTextRapidBlueprint
from rapidata.api_client.models.line_rapid_blueprint import LineRapidBlueprint
from rapidata.api_client.models.locate_rapid_blueprint import LocateRapidBlueprint
from rapidata.api_client.models.named_entity_rapid_blueprint import NamedEntityRapidBlueprint
from rapidata.api_client.models.polygon_rapid_blueprint import PolygonRapidBlueprint
from rapidata.api_client.models.transcription_rapid_blueprint import TranscriptionRapidBlueprint
from pydantic import StrictStr, Field
from typing import Union, List, Set, Optional, Dict
from typing_extensions import Literal, Self

SIMPLEWORKFLOWCONFIGMODELBLUEPRINT_ONE_OF_SCHEMAS = ["AttachCategoryRapidBlueprint", "BoundingBoxRapidBlueprint", "CompareRapidBlueprint", "FreeTextRapidBlueprint", "LineRapidBlueprint", "LocateRapidBlueprint", "NamedEntityRapidBlueprint", "PolygonRapidBlueprint", "TranscriptionRapidBlueprint"]

class SimpleWorkflowConfigModelBlueprint(BaseModel):
    """
    The blueprint to use when creating rapids for this workflow.  The blueprint determines what kind of rapids will be created and what data they will contain.
    """
    # data type: TranscriptionRapidBlueprint
    oneof_schema_1_validator: Optional[TranscriptionRapidBlueprint] = None
    # data type: PolygonRapidBlueprint
    oneof_schema_2_validator: Optional[PolygonRapidBlueprint] = None
    # data type: NamedEntityRapidBlueprint
    oneof_schema_3_validator: Optional[NamedEntityRapidBlueprint] = None
    # data type: LocateRapidBlueprint
    oneof_schema_4_validator: Optional[LocateRapidBlueprint] = None
    # data type: LineRapidBlueprint
    oneof_schema_5_validator: Optional[LineRapidBlueprint] = None
    # data type: FreeTextRapidBlueprint
    oneof_schema_6_validator: Optional[FreeTextRapidBlueprint] = None
    # data type: CompareRapidBlueprint
    oneof_schema_7_validator: Optional[CompareRapidBlueprint] = None
    # data type: AttachCategoryRapidBlueprint
    oneof_schema_8_validator: Optional[AttachCategoryRapidBlueprint] = None
    # data type: BoundingBoxRapidBlueprint
    oneof_schema_9_validator: Optional[BoundingBoxRapidBlueprint] = None
    actual_instance: Optional[Union[AttachCategoryRapidBlueprint, BoundingBoxRapidBlueprint, CompareRapidBlueprint, FreeTextRapidBlueprint, LineRapidBlueprint, LocateRapidBlueprint, NamedEntityRapidBlueprint, PolygonRapidBlueprint, TranscriptionRapidBlueprint]] = None
    one_of_schemas: Set[str] = { "AttachCategoryRapidBlueprint", "BoundingBoxRapidBlueprint", "CompareRapidBlueprint", "FreeTextRapidBlueprint", "LineRapidBlueprint", "LocateRapidBlueprint", "NamedEntityRapidBlueprint", "PolygonRapidBlueprint", "TranscriptionRapidBlueprint" }

    model_config = ConfigDict(
        validate_assignment=True,
        protected_namespaces=(),
    )


    discriminator_value_class_map: Dict[str, str] = {
    }

    def __init__(self, *args, **kwargs) -> None:
        if args:
            if len(args) > 1:
                raise ValueError("If a position argument is used, only 1 is allowed to set `actual_instance`")
            if kwargs:
                raise ValueError("If a position argument is used, keyword arguments cannot be used.")
            super().__init__(actual_instance=args[0])
        else:
            super().__init__(**kwargs)

    @field_validator('actual_instance')
    def actual_instance_must_validate_oneof(cls, v):
        instance = SimpleWorkflowConfigModelBlueprint.model_construct()
        error_messages = []
        match = 0
        # validate data type: TranscriptionRapidBlueprint
        if not isinstance(v, TranscriptionRapidBlueprint):
            error_messages.append(f"Error! Input type `{type(v)}` is not `TranscriptionRapidBlueprint`")
        else:
            match += 1
        # validate data type: PolygonRapidBlueprint
        if not isinstance(v, PolygonRapidBlueprint):
            error_messages.append(f"Error! Input type `{type(v)}` is not `PolygonRapidBlueprint`")
        else:
            match += 1
        # validate data type: NamedEntityRapidBlueprint
        if not isinstance(v, NamedEntityRapidBlueprint):
            error_messages.append(f"Error! Input type `{type(v)}` is not `NamedEntityRapidBlueprint`")
        else:
            match += 1
        # validate data type: LocateRapidBlueprint
        if not isinstance(v, LocateRapidBlueprint):
            error_messages.append(f"Error! Input type `{type(v)}` is not `LocateRapidBlueprint`")
        else:
            match += 1
        # validate data type: LineRapidBlueprint
        if not isinstance(v, LineRapidBlueprint):
            error_messages.append(f"Error! Input type `{type(v)}` is not `LineRapidBlueprint`")
        else:
            match += 1
        # validate data type: FreeTextRapidBlueprint
        if not isinstance(v, FreeTextRapidBlueprint):
            error_messages.append(f"Error! Input type `{type(v)}` is not `FreeTextRapidBlueprint`")
        else:
            match += 1
        # validate data type: CompareRapidBlueprint
        if not isinstance(v, CompareRapidBlueprint):
            error_messages.append(f"Error! Input type `{type(v)}` is not `CompareRapidBlueprint`")
        else:
            match += 1
        # validate data type: AttachCategoryRapidBlueprint
        if not isinstance(v, AttachCategoryRapidBlueprint):
            error_messages.append(f"Error! Input type `{type(v)}` is not `AttachCategoryRapidBlueprint`")
        else:
            match += 1
        # validate data type: BoundingBoxRapidBlueprint
        if not isinstance(v, BoundingBoxRapidBlueprint):
            error_messages.append(f"Error! Input type `{type(v)}` is not `BoundingBoxRapidBlueprint`")
        else:
            match += 1
        if match > 1:
            # more than 1 match
            raise ValueError("Multiple matches found when setting `actual_instance` in SimpleWorkflowConfigModelBlueprint with oneOf schemas: AttachCategoryRapidBlueprint, BoundingBoxRapidBlueprint, CompareRapidBlueprint, FreeTextRapidBlueprint, LineRapidBlueprint, LocateRapidBlueprint, NamedEntityRapidBlueprint, PolygonRapidBlueprint, TranscriptionRapidBlueprint. Details: " + ", ".join(error_messages))
        elif match == 0:
            # no match
            raise ValueError("No match found when setting `actual_instance` in SimpleWorkflowConfigModelBlueprint with oneOf schemas: AttachCategoryRapidBlueprint, BoundingBoxRapidBlueprint, CompareRapidBlueprint, FreeTextRapidBlueprint, LineRapidBlueprint, LocateRapidBlueprint, NamedEntityRapidBlueprint, PolygonRapidBlueprint, TranscriptionRapidBlueprint. Details: " + ", ".join(error_messages))
        else:
            return v

    @classmethod
    def from_dict(cls, obj: Union[str, Dict[str, Any]]) -> Self:
        return cls.from_json(json.dumps(obj))

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Returns the object represented by the json string"""
        instance = cls.model_construct()
        error_messages = []
        match = 0

        # deserialize data into TranscriptionRapidBlueprint
        try:
            instance.actual_instance = TranscriptionRapidBlueprint.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into PolygonRapidBlueprint
        try:
            instance.actual_instance = PolygonRapidBlueprint.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into NamedEntityRapidBlueprint
        try:
            instance.actual_instance = NamedEntityRapidBlueprint.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into LocateRapidBlueprint
        try:
            instance.actual_instance = LocateRapidBlueprint.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into LineRapidBlueprint
        try:
            instance.actual_instance = LineRapidBlueprint.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into FreeTextRapidBlueprint
        try:
            instance.actual_instance = FreeTextRapidBlueprint.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into CompareRapidBlueprint
        try:
            instance.actual_instance = CompareRapidBlueprint.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into AttachCategoryRapidBlueprint
        try:
            instance.actual_instance = AttachCategoryRapidBlueprint.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into BoundingBoxRapidBlueprint
        try:
            instance.actual_instance = BoundingBoxRapidBlueprint.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))

        if match > 1:
            # more than 1 match
            raise ValueError("Multiple matches found when deserializing the JSON string into SimpleWorkflowConfigModelBlueprint with oneOf schemas: AttachCategoryRapidBlueprint, BoundingBoxRapidBlueprint, CompareRapidBlueprint, FreeTextRapidBlueprint, LineRapidBlueprint, LocateRapidBlueprint, NamedEntityRapidBlueprint, PolygonRapidBlueprint, TranscriptionRapidBlueprint. Details: " + ", ".join(error_messages))
        elif match == 0:
            # no match
            raise ValueError("No match found when deserializing the JSON string into SimpleWorkflowConfigModelBlueprint with oneOf schemas: AttachCategoryRapidBlueprint, BoundingBoxRapidBlueprint, CompareRapidBlueprint, FreeTextRapidBlueprint, LineRapidBlueprint, LocateRapidBlueprint, NamedEntityRapidBlueprint, PolygonRapidBlueprint, TranscriptionRapidBlueprint. Details: " + ", ".join(error_messages))
        else:
            return instance

    def to_json(self) -> str:
        """Returns the JSON representation of the actual instance"""
        if self.actual_instance is None:
            return "null"

        if hasattr(self.actual_instance, "to_json") and callable(self.actual_instance.to_json):
            return self.actual_instance.to_json()
        else:
            return json.dumps(self.actual_instance)

    def to_dict(self) -> Optional[Union[Dict[str, Any], AttachCategoryRapidBlueprint, BoundingBoxRapidBlueprint, CompareRapidBlueprint, FreeTextRapidBlueprint, LineRapidBlueprint, LocateRapidBlueprint, NamedEntityRapidBlueprint, PolygonRapidBlueprint, TranscriptionRapidBlueprint]]:
        """Returns the dict representation of the actual instance"""
        if self.actual_instance is None:
            return None

        if hasattr(self.actual_instance, "to_dict") and callable(self.actual_instance.to_dict):
            return self.actual_instance.to_dict()
        else:
            # primitive type
            return self.actual_instance

    def to_str(self) -> str:
        """Returns the string representation of the actual instance"""
        return pprint.pformat(self.model_dump())





from abc import ABC, abstractmethod
from typing import Any, Mapping

class Referee(ABC):
    """
    The referee defines when a rapid is considered complete. 
    """
    @abstractmethod
    def _to_dict(self) -> Mapping[str, str | int | float]:
        """
        Convert the referee to a referee configuration dict.
        """
        pass

    @abstractmethod
    def _to_model(self) -> Any:
        """
        Convert the referee to a referee configuration model.
        """
        pass



from rapidata.rapidata_client.selection._base_selection import RapidataSelection
from rapidata.api_client.models.demographic_selection import DemographicSelection as DemographicSelectionModel


class DemographicSelection(RapidataSelection):
    """Demographic selection class.
    
    This is used to ask demographic questions in an order.    

    The keys will select the rapids based on the confidence we already saved for each user.

    If the confidence is high, the users will be selected to solve the rapids with lower probability.

    Args:
        keys (list[str]): List of keys for the demographic rapids to be shown. As an example: "age"
        max_rapids (int): The maximum number of rapids to run.\n
            Allows to provide more keys, in case some of the earlier ones are not selected because of high confidence."""

    def __init__(self, keys: list[str], max_rapids: int):
        self.keys = keys
        self.max_rapids = max_rapids

    def _to_model(self):
        return DemographicSelectionModel(_t="DemographicSelection", keys=self.keys, maxRapids=self.max_rapids)
